c     Copyright (C)
c     2021 Mésotar
c     
c     This file must be used under the terms of the CeCILL license.
c     This source file is licensed as described in the file COPYING, which
c     you should have received as part of this distribution.  The terms
c     are also available at
c     http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.txt
c
      subroutine intersection_with_box_from_outside(debug,dim,check1,
     &     box_dimensions,x0,u0,
     &     intersection_found,x_int,d_x0xint,n_int)
      implicit none
      include 'max.inc'
c
c     Purpose: to find position "x_int" of the intersection between a line
c     starting at x0 (from outside a box), in direction u0, and the walls of the box
c
c     Input:
c       + dim: space dimension
c       + check1: true if the input position "x0" has to be checked into the box
c       + box_dimensions(i,j):
c         - i: index of the dimension (1:x, 2:y, 3:z)
c         - j: index of limit (1: lower, 2: higher)
c       + x0: starting position
c       + u0: propagation direction
c
c     Output:
c       + intersection_found: true if an intersection exists
c       + x_int: intersection position
c       + d_x0xint: distance between x0 and x_int
c       + n_int: normal to the surface at x_int
c     
c     I/O
      logical debug
      integer dim
      logical check1
      double precision box_dimensions(1:Ndim_mx,1:2)
      double precision x0(1:Ndim_mx)
      double precision u0(1:Ndim_mx)
      logical intersection_found
      double precision x_int(1:Ndim_mx)
      double precision d_x0xint
      double precision n_int(1:Ndim_mx)
c     temp
      integer idim,jdim,i,j,k,idim_int,j_int
      logical int_found(1:2*Ndim_mx),int_found_tmp
      double precision int_positions(1:2*Ndim_mx,1:Ndim_mx)
      double precision length(1:2*Ndim_mx)
      logical in_box
      integer int_index
      double precision x_int_tmp(1:Ndim_mx)
      double precision min_length
      logical on_surface
      integer surface_index1
      integer surface_index2
      double precision n_wall(1:Ndim_mx)
      double precision product
      double precision x2(1:Ndim_mx)
      logical identical
c     label
      character*(Nchar_mx) label
      label='subroutine intersection_with_box_from_outside'

c     security
      if (check1) then
         call is_in_box(dim,x0,box_dimensions,in_box)
c     checking whether or not position "x0" is located on a surface
         call is_on_surface(dim,x0,box_dimensions,
     &        on_surface,surface_index1,surface_index2)
         if ((in_box).and.(.not.on_surface)) then
            call error(label)
            write(*,*) 'x0=',(x0(i),i=1,dim)
            write(*,*) 'is inside the box'
            stop
         endif
      endif
         
c     If position "x0" is located on a surface, the propagation direction
c     has to be checked against the normal to the surface at x0
      if (on_surface) then
         call normal_at_wall(dim,surface_index1,surface_index2,n_wall)
         call scalar_product(dim,u0,n_wall,product)
c     if product>0, x0 is the intersection position
         if (product.gt.0.0D+0) then
            intersection_found=.true.
            call duplicate_vector(dim,x0,x_int)
            d_x0xint=0.0D+0
            call duplicate_vector(dim,n_wall,n_int)
            goto 666
         endif
      endif                     ! on_surface

c     looking for an intersection
      i=0
      min_length=-1.0D+0
      intersection_found=.false.
      do idim=1,dim
         do j=1,2
            i=i+1
            call intersection_xplane(dim,x0,u0,idim,
     &           box_dimensions(idim,j),int_found_tmp,
     &           x_int_tmp,length(i))
c     in this case, the intersection with the plane also has to be
c     located in a given rectangle
            if (int_found_tmp) then
               int_found(i)=.true. ! default value
               do jdim=1,dim
                  if (jdim.ne.idim) then
                     if ((x_int_tmp(jdim).lt.box_dimensions(jdim,1))
     &                    .or.(x_int_tmp(jdim).gt.
     &                    box_dimensions(jdim,2))) then
                        int_found(i)=.false.
                     endif
                  endif         ! jdim.ne.idim
               enddo            ! jdim
            else
               int_found(i)=.false.
            endif               ! int_found_tmp            
            do k=1,dim
               int_positions(i,k)=x_int_tmp(k)
            enddo ! k
c     First time an intersection is found
            if ((.not.intersection_found)
     &           .and.(int_found(i))
     &           .and.(length(i).gt.0.0D+0)) then
c     remove case when x0 is on a surface
               if ((on_surface).and.
     &              (surface_index1.eq.idim).and.
     &              (surface_index2.eq.j)) then
                  goto 123 ! do nothing
               else
                  intersection_found=.true.
                  int_index=i
                  idim_int=idim
                  j_int=j
                  min_length=length(i)
               endif
 123           continue
            endif
c     Subsequent intersections
            if ((intersection_found)
     &           .and.(int_found(i))
     &           .and.(length(i).gt.0.0D+0)
     &           .and.(length(i).lt.min_length)) then
c     remove case when x0 is on a surface
               if ((on_surface).and.
     &              (surface_index1.eq.idim).and.
     &              (surface_index2.eq.j)) then
                  goto 124 ! do nothing
               else
                  int_index=i
                  idim_int=idim
                  j_int=j
                  min_length=length(i)
               endif
 124           continue
            endif
         enddo ! j
      enddo ! idim

c     get intersection position (if it has been found)
      if (intersection_found) then
         do k=1,dim
            x_int(k)=int_positions(int_index,k)
         enddo ! k
         d_x0xint=min_length
         call normal_at_wall(dim,idim_int,j_int,n_int)
      endif

 666  continue
c     double-check
      if (intersection_found) then
         if (d_x0xint.lt.0.0D+0) then
            call error(label)
            write(*,*) 'd_x0xint=',d_x0xint
            stop
         endif
         call direct_flight(dim,x0,u0,d_x0xint,x2)
         call identical_vectors(dim,x_int,x2,identical)
         if (.not.identical) then
            call error(label)
            write(*,*) 'x_int=',(x_int(i),i=1,dim)
            write(*,*) 'x0+d_x0_xint*u0=',(x2(i),i=1,dim)
            write(*,*) 'd_x0xint=',d_x0xint
            stop
         endif
      endif                     ! intersection_found

      return
      end
      


      subroutine intersection_with_box_from_inside(debug,dim,check1,
     &     box_dimensions,x0,u0,
     &     intersection_found,x_int,d_x0xint,n_int,errcode)
      implicit none
      include 'max.inc'
c
c     Purpose: to find position "x_int" of the intersection between a line
c     starting at x0 (from inside a box), in direction u0, and the walls of the box
c
c     Input:
c       + dim: space dimension
c       + check1: true if the input position "x0" has to be checked into the box
c       + box_dimensions(i,j):
c         - i: index of the dimension (1:x, 2:y, 3:z)
c         - j: index of limit (1: lower, 2: higher)
c       + x0: starting position
c       + u0: propagation direction
c
c     Output:
c       + intersection_found: true if an intersection exists
c       + x_int: intersection position
c       + d_x0xint: distance between x0 and x_int
c       + n_int: normal to the surface at x_int
c       + errcode: error code (0: no error; 1: something wrong happened)
c     
c     I/O
      logical debug
      integer dim
      logical check1
      double precision box_dimensions(1:Ndim_mx,1:2)
      double precision x0(1:Ndim_mx)
      double precision u0(1:Ndim_mx)
      logical intersection_found
      double precision x_int(1:Ndim_mx)
      double precision d_x0xint
      double precision n_int(1:Ndim_mx)
      integer errcode
c     temp
      integer idim,i,j,k,idim_int,j_int
      logical int_found(1:2*Ndim_mx)
      double precision int_positions(1:2*Ndim_mx,1:Ndim_mx)
      double precision length(1:2*Ndim_mx)
      logical in_box
      integer int_index
      double precision x_int_tmp(1:Ndim_mx)
      double precision min_length
      logical on_surface
      integer surface_index1
      integer surface_index2
      double precision n_wall(1:Ndim_mx)
      double precision product
      double precision x2(1:Ndim_mx)
      logical identical
c     label
      character*(Nchar_mx) label
      label='subroutine intersection_with_box_from_inside'

      errcode=0      
c     security
      if (check1) then
         call is_in_box(dim,x0,box_dimensions,in_box)
         if (.not.in_box) then
            call error(label)
            write(*,*) 'x0=',(x0(i),i=1,dim)
            write(*,*) 'is not in the geometry'
            stop
         endif
      endif
c     checking whether or not position "x0" is located on a surface
      call is_on_surface(dim,x0,box_dimensions,
     &     on_surface,surface_index1,surface_index2)
c     If position "x0" is located on a surface, the propagation direction
c     has to be checked against the normal to the surface at x0
      if (on_surface) then
         call normal_at_wall(dim,surface_index1,surface_index2,n_wall)
         call scalar_product(dim,u0,n_wall,product)
         if (product.lt.0.0D+0) then
            errcode=1
            goto 666
         endif
      endif                     ! on_surface

c     looking for an intersection
      i=0
      min_length=-1.0D+0
      intersection_found=.false.
      do idim=1,dim
         do j=1,2
            i=i+1
            call intersection_xplane(dim,x0,u0,idim,
     &           box_dimensions(idim,j),int_found(i),
     &           x_int_tmp,length(i))
            do k=1,dim
               int_positions(i,k)=x_int_tmp(k)
            enddo ! k
c     First time an intersection is found
            if ((.not.intersection_found)
     &           .and.(int_found(i))
     &           .and.(length(i).gt.0.0D+0)) then
c     remove case when x0 is on a surface
               if ((on_surface).and.
     &              (surface_index1.eq.idim).and.
     &              (surface_index2.eq.j)) then
                  goto 123 ! do nothing
               else
                  intersection_found=.true.
                  int_index=i
                  idim_int=idim
                  j_int=j
                  min_length=length(i)
               endif
 123           continue
            endif
c     Subsequent intersections
            if ((intersection_found)
     &           .and.(int_found(i))
     &           .and.(length(i).gt.0.0D+0)
c     &           .and.(length(i).gt.1.0D-8)
     &           .and.(length(i).lt.min_length)) then
c     remove case when x0 is on a surface
               if ((on_surface).and.
     &              (surface_index1.eq.idim).and.
     &              (surface_index2.eq.j)) then
                  goto 124 ! do nothing
               else
                  int_index=i
                  idim_int=idim
                  j_int=j
                  min_length=length(i)
               endif
 124           continue
            endif
         enddo ! j
      enddo ! idim

c     get intersection position (if it has been found)
      if (intersection_found) then
         do k=1,dim
            x_int(k)=int_positions(int_index,k)
         enddo ! k
         d_x0xint=min_length
c         d_x0xint=length(int_index)
         call normal_at_wall(dim,idim_int,j_int,n_int)
      else
c     If no intersection was found, there was an error somewhere
         call error(label)
         write(*,*) 'Intersection with walls not found'
         write(*,*) 'x0=',(x0(k),k=1,dim)
         write(*,*) 'u0=',(u0(k),k=1,dim)
         do i=1,dim
            write(*,*) 'box_dimensions(',i,')=',
     &           (box_dimensions(i,j),j=1,2)
         enddo ! i
         stop
      endif

 321  continue
c     double-check
      if (d_x0xint.lt.0.0D+0) then
         call error(label)
         write(*,*) 'd_x0xint=',d_x0xint
         stop
      endif
      call direct_flight(dim,x0,u0,d_x0xint,x2)
      call identical_vectors(dim,x_int,x2,identical)
      if (.not.identical) then
         call error(label)
         write(*,*) 'x_int=',(x_int(i),i=1,dim)
         write(*,*) 'x0+d_x0_xint*u0=',(x2(i),i=1,dim)
         stop
      endif

 666  continue
      
      return
      end



      subroutine normal_at_wall(dim,idim_int,j_int,n_wall)
      implicit none
      include 'max.inc'
c
c     Purpose: to identify the vector normal to a wall
c
c     Input:
c       + dim: space dimension
c       + (idim_int,j_int): indexes that identify the surface
c
c     Output:
c       + n_wall: normal to the surface (pointing inside the box)
c
c     I/O
      integer dim
      integer idim_int,j_int
      double precision n_wall(1:Ndim_mx)
c     temp
      integer i
      double precision plus_one,minus_one
      double precision a
c     label
      character*(Nchar_mx) label
      label='subroutine normal_at_wall'

      plus_one=1.0D+0
      minus_one=-1.0D+0
      do i=1,dim
         n_wall(i)=0.0D+0
      enddo
      if (j_int.eq.1) then
         a=plus_one
      else if (j_int.eq.2) then
         a=minus_one
      else
         call error(label)
         write(*,*) 'j_int=',j_int
         stop
      endif
      n_wall(idim_int)=a

      return
      end



      subroutine intersection_xplane(dim,x0,u0,idim,plane,
     &     intersection_found,x_int,length)
      implicit none
      include 'max.inc'
c
c     Purpose: to find position "x_int" of the intersection between a line
c     starting at x0, in direction u0, and a plane defined by x_idim=plane
c
c     Input:
c       + dim: space dimension
c       + x0: starting position
c       + u0: propagation direction
c       + idim: dimension index that defines the type of plane
c         - idim=1: plane defined by x=plane=constant
c         - idim=2: plane defined by y=plane=constant
c         - idim=3: plane defined by z=plane=constant
c       + plane: constant that defines the x_idim=xplane plane
c     
c     Output:
c       + intersection_found: true if an intersection exists
c       + x_int: intersection position
c       + length: distance from x0 to x_int (positive or negative value)
c
c     I/O
      integer dim
      double precision x0(1:Ndim_mx)
      double precision u0(1:Ndim_mx)
      integer idim
      double precision plane
      logical intersection_found
      double precision x_int(1:Ndim_mx)
c     temp
      double precision length
c     label
      character*(Nchar_mx) label
      label='subroutine intersection_xplane'

c     security
      if ((idim.lt.1).or.(idim.gt.dim)) then
         call error(label)
         write(*,*) 'parameter idim=',idim
         stop
      endif

c     special case of a propagation direction parallel to the plane: no intersection is possible
      if (u0(idim).eq.0.0D+0) then
         intersection_found=.false.
c     general case: it is possible to find an intersection !
      else
         intersection_found=.true.
         length=(plane-x0(idim))/u0(idim)
         call direct_flight(dim,x0,u0,length,x_int)
c     clamp component index 'idim' to value of 'plane'
         x_int(idim)=plane
      endif

      return
      end



      subroutine get_direction(dim,x1,x2,u)
      implicit none
      include 'max.inc'
c
c     Purpose: to compute the direction from x1 to x2
c
c     Input:
c       + dim: space dimension
c       + x1: first position
c       + x2: second position
c
c     Output:
c       + u: direction from x1 to x2
c
c     I/O
      integer dim
      double precision x1(1:Ndim_mx)
      double precision x2(1:Ndim_mx)
      double precision u(1:Ndim_mx)
c     temp
      double precision v(1:Ndim_mx)
      logical identical
      integer i
c     label
      character*(Nchar_mx) label
      label='subroutine get_direction'

c     security: vectors "x1" and "x2" must be different
      call identical_vectors(dim,x1,x2,identical)
      if (identical) then
         call error(label)
         write(*,*) 'x1=',(x1(i),i=1,dim)
         write(*,*) 'x2=',(x2(i),i=1,dim)
         write(*,*) 'are identical'
         stop
      else
         call substract_vectors(dim,x2,x1,v) ! v=x2-x1
         call normalize_vector(dim,v,u) ! u=v/|v|
      endif

      return
      end


      
      subroutine direct_flight(dim,x1,u,length,x2)
      implicit none
      include 'max.inc'
c
c     Purpose: to compute x2=x1+length*u
c
c     Input:
c       + dim: space dimension
c       + x1: starting position
c       + u: flight direction
c       + length: flight length
c     
c     Output:
c       + x2=x1+length*u
c
c     I/O
      integer dim
      double precision x1(1:Ndim_mx)
      double precision u(1:Ndim_mx)
      double precision length
      double precision x2(1:Ndim_mx)
c     temp
      double precision flight(1:Ndim_mx)
c     label
      character*(Nchar_mx) label
      label='subroutine direct_flight'

      call scalar_vector(dim,length,u,flight)
      call add_vectors(dim,x1,flight,x2)

      return
      end


      
      subroutine wall_surface(dim,box_dimensions,idx1,idx2,surface)
      implicit none
      include 'max.inc'
c
c     Purpose: to compute the surface of a given wall
c
c     Input:
c       + dim: space dimension
c       + box_dimensions(i,j):
c         - i: index of the dimension (1:x, 2:y, 3:z)
c         - j: index of limit (1: lower, 2: higher)
c       + (idx1,idx2): indexes that define the wall
c
c     Output:
c       + surface: value of the area
c
c     I/O
      integer dim
      double precision box_dimensions(1:Ndim_mx,1:2)
      integer idx1
      integer idx2
      double precision surface
c     temp
      integer i
      double precision d1,d2,d
      logical d1_set
c     label
      character*(Nchar_mx) label
      label='subroutine wall_surface'

      d1_set=.false.
      do i=1,dim
         if (i.ne.idx1) then
            d=box_dimensions(i,2)-box_dimensions(i,1)
            if (d.le.0.0D+0) then
               call error(label)
               write(*,*) 'd=',d
               stop
            endif
            if (.not.d1_set) then
               d1=d
               d1_set=.true.
            else
               d2=d
            endif
         endif
      enddo
      surface=d1*d2

      return
      end


      
      subroutine box_total_surface(dim,box_dimensions,surface)
      implicit none
      include 'max.inc'
c
c     Purpose: to compute the total surface of the box
c
c     Input:
c       + dim: space dimension
c       + box_dimensions(i,j):
c         - i: index of the dimension (1:x, 2:y, 3:z)
c         - j: index of limit (1: lower, 2: higher)
c
c     Output:
c       + surface: total surface of the box
c
c     I/O
      integer dim
      double precision box_dimensions(1:Ndim_mx,1:2)
      double precision surface
c     temp
      integer i,j
      double precision S
c     label
      character*(Nchar_mx) label
      label='subroutine box_total_surface'

      surface=0.0D+0
      do i=1,dim
         do j=1,2
            call wall_surface(dim,box_dimensions,i,j,S)
            surface=surface+S
         enddo ! j
      enddo ! i

      return
      end



      subroutine box_volume(dim,box_dimensions,volume)
      implicit none
      include 'max.inc'
c
c     Purpose: to compute the volume of the box
c
c     Input:
c       + dim: space dimension
c       + box_dimensions(i,j):
c         - i: index of the dimension (1:x, 2:y, 3:z)
c         - j: index of limit (1: lower, 2: higher)
c
c     Output:
c       + volume: volume of the box
c
c     I/O
      integer dim
      double precision box_dimensions(1:Ndim_mx,1:2)
      double precision volume
c     temp
      integer i
      double precision d
c     label
      character*(Nchar_mx) label
      label='subroutine box_volume'

      do i=1,dim
         d=box_dimensions(i,2)-box_dimensions(i,1)
         if (d.le.0.0D+0) then
            call error(label)
            write(*,*) 'd=',d
            stop
         endif
         if (i.eq.1) then
            volume=d
         else
            volume=volume*d
         endif
      enddo ! i

      return
      end
