c     Copyright (C)
c     2021 Mésotar
c     
c     This file must be used under the terms of the CeCILL license.
c     This source file is licensed as described in the file COPYING, which
c     you should have received as part of this distribution.  The terms
c     are also available at
c     http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.txt
c
      subroutine random_gen(tmp)
      implicit none
      include 'max.inc'
c
c     Main pseudo-random number generator routine
c     (the one that must be called)
c
c     Output:
c       + tmp: pseudo-random number in [0,1]
c
c     I/O
      double precision tmp
c     temp
      integer n,t
c     label
      character*(Nchar_mx) label
      label='subroutine random_gen'

      n=1
      t=0
 10   continue
      call zufall(n,tmp)
      if (tmp.eq.0.or.tmp.eq.1) then
         t=t+1
         if (t.gt.10) then
            call error(label)
            write(*,*) 'Too many zeroes'
            stop
         endif
         goto 10
      endif

      return
      end



      subroutine uniform(val1,val2,alpha)
      implicit none
      include 'param.inc'
      include 'max.inc'
c
c     Purpose: to return a value between "val1" and "val2", chosen uniformally
c
c     Input:
c       + val1: lower limit
c       + val2: higher limit
c
c     Output:
c       + alpha: random value chosen uniformally in [val1,val2]
c
c     I/O
      double precision val1,val2,alpha
c     temp
      double precision R
c     label
      character*(Nchar_mx) label
      label='subroutine uniform'

      if (val2.le.val1) then
         call error(label)
         write(*,*) 'val2=',val2
         write(*,*) '< val1=',val1
         stop
      endif
      call random_gen(R)
      alpha=val1+(val2-val1)*R

      return
      end



      subroutine choose_integer(N1,N2,i)
      implicit none
      include 'max.inc'
c
c     Purpose: to choose an integer between N1 and N2 (included)
c     on a random uniform basis
c
c     Input:
c       + N1: minimum value the integer can take
c       + N2: maximum value the integer can take
c
c     Output:
c       + i: integer value chosen uniformally between N1 and N2
c
c     I/O
      integer N1,N2,i
c     temp
      double precision R
c     label
      character*(Nchar_mx) label
      label='subroutine choose_integer'

      call random_gen(R)
      i=N1+int((N2-N1)*R)

      return
      end

      
      
      subroutine choose_emission_direction_on_surface(dim,normal,
     &     u,pdf_u)
      implicit none
      include 'max.inc'
c
c     Purpose: to choose an emission direction from a position on a surface
c     according to pdf(u)=|u.n|/pi
c
c     Input:
c       + dim: space dimension
c       + normal: normal to the surface @ emission position
c
c     Output:
c       + u: direction
c       + pdf_u: corresponding pdf
c       
c     I/O
      integer dim
      double precision normal(1:Ndim_mx)
      double precision u(1:Ndim_mx)
      double precision pdf_u
c     temp
      integer i
      double precision j1(1:Ndim_mx)
      double precision theta,phi
      double precision pdf_theta,pdf_phi
      double precision v(1:Ndim_mx)
      double precision product_normalj1
      double precision  product_un
      double precision normals(1:Ndim_mx)
c     label
      character*(Nchar_mx) label
      label='subroutine choose_emission_direction_on_surface'

      call vector_j1(dim,normal,j1)
      call choose_theta_surface(theta,pdf_theta)
      call choose_phi_uniform(phi,pdf_phi)

c     First rotation: vector "normal" is rotated by "theta" around "j1" -> vector "v"
      call rotation(dim,normal,theta,j1,v)
      call scalar_product(dim,normal,j1,product_normalj1)
      if (dabs(product_normalj1).gt.1.0D-8) then
         call error(label)
         write(*,*) 'normal=',(normal(i),i=1,dim)
         write(*,*) 'j1=',(j1(i),i=1,dim)
         write(*,*) '|normal.j1|=',product_normalj1
         write(*,*) 'should be null'
         call cart2spher(normal,normals)
         write(*,*) 'normals=',(normals(i),i=1,dim)
         stop
      endif
c     Second rotation: vector "v" is rotated by "phi" around "normal" -> vector "u"
      call rotation(dim,v,phi,normal,u)
c     compute pdf_u(u)
      pdf_u=pdf_theta*pdf_phi
c     double check
      call scalar_product(dim,normal,u,product_un)
      if (dabs(product_un-dcos(theta)).gt.1.0D-8) then
         call error(label)
         write(*,*) 'cos(',theta,')=',dcos(theta)
         write(*,*) '|u.n|=',product_un
         write(*,*) 'normal=',(normal(i),i=1,dim)
         write(*,*) 'j1=',(j1(i),i=1,dim)
         write(*,*) '|normal.j1|=',product_normalj1
         stop
      endif

      return
      end



      subroutine choose_theta_surface(theta,pdf_theta)
      implicit none
      include 'max.inc'
      include 'param.inc'
c
c     Purpose: to pick up an emission angle [0,pi/2] from a boundary
c     according to pdf(w)=|u.n|/pi
c
c     Outputs:
c       + theta: emission angle [0,pi/2] between the normal and a direction
c       + pdf_theta: probability density function corresponding to the chosen angle
c

c     output
      double precision theta
      double precision pdf_theta
c     temp
      double precision R
c     label
      character*(Nchar_mx) label
      label='subroutine choose_theta_surface'

      call random_gen(R)
      theta=dacos(dsqrt(R))
      pdf_theta=2.0D+0*dcos(theta)
c     Tests
      if ((theta.lt.0.0D+0).or.(theta.gt.pi/2.0D+0)) then
         call error(label)
         write(*,*) 'theta=',theta
         write(*,*) 'is not between 0 and pi/2'
         write(*,*) 'R=',R
         stop
      endif
c     Tests

      return
      end



      subroutine choose_emission_direction_in_volume(dim,u,pdf_u)
      implicit none
      include 'max.inc'
c
c     Purpose: to chosse a volumic emission direction (uniformally over 4*pi)
c
c     Input:
c       + dim: space dimension
c
c     Output:
c       + u: emission direction
c       + pdf_u: associated probability density function
c
c     I/O
      integer dim
      double precision u(1:Ndim_mx)
      double precision pdf_u
c     temp
      double precision u_spherical(1:Ndim_mx)
      double precision pdf_theta
      double precision pdf_phi
      logical verif
c     label
      character*(Nchar_mx) label
      label='subroutine choose_emission_direction_in_volume'
      
      u_spherical(1)=1.0D+0
      call choose_theta_volume(u_spherical(2),pdf_theta)
      call choose_phi_uniform(u_spherical(3),pdf_phi)

      verif=.true.
      call spher2cart(label,verif,u_spherical,u)
      pdf_u=pdf_theta*pdf_phi

      return
      end



      subroutine choose_phi_uniform(phi,pdf_phi)
      implicit none
      include 'max.inc'
c
c     Purpose: to choose an emission angle (volumic emission) phi
c     uniformally in the [0,2*pi] range
c
c     Output:
c       + phi: value of phi in the [0,2*pi] range
c       + pdf_phi: associated probability density function
c
c     I/O
      double precision phi
      double precision pdf_phi
c     temp
      double precision R
      double precision pi
c     label
      character*(Nchar_mx) label
      label='subroutine choose_phi_uniform'

      pi=4.0D+0*datan(1.0D+0)

      call random_gen(R)
      phi=2.0D+0*pi*R
      pdf_phi=1.0D+0/(2.0D+0*pi)

      return
      end



      subroutine choose_theta_volume(theta,pdf_theta)
      implicit none
      include 'max.inc'
c
c     Purpose: to choose an emission angle (volumic emission) theta
c     uniformally in the [-pi/2,pi/2] range
c
c     Output:
c       + theta: value of theta in the [-pi/2,pi/2] range
c       + pdf_theta: associated probability density function
c
c     I/O
      double precision theta
      double precision pdf_theta
c     temp
      double precision R
c     label
      character*(Nchar_mx) label
      label='subroutine choose_theta_volume'

      call random_gen(R)
      theta=dasin(2.0D+0*R-1.0D+0)
      pdf_theta=0.5D+0

      return
      end



      subroutine choose_direction_in_solid_angle(dim,
     &     theta_min,theta_max,phi_min,phi_max,u0)
      implicit none
      include 'max.inc'
c     
c     Purpose: to sample a direction in a given solid angle
c     defined by its latitude / longitude limits
c     
c     Input:
c       + dim: dimension of space
c       + theta_min: minimum latitude of the solid angle [rad] (-pi/2,pi/2)
c       + theta_max: maximum latitude of the solid angle [rad] (-pi/2,pi/2)
c       + phi_min: minimum longitude of the solid angle [rad] (0,2*pi)
c       + phi_max: maximum longitude of the solid angle [rad] (0,2*pi)
c     
c     Output:
c       + u0: direction
c     
c     I/O
      integer dim
      double precision theta_min,theta_max
      double precision phi_min,phi_max
      double precision u0(1:Ndim_mx)
c     temp
      double precision theta,phi
      double precision pdf_theta,pdf_phi
c     label
      character*(Nchar_mx) label
      label='subroutine choose_theta_in_interval'

      call choose_theta_in_interval(theta_min,theta_max,
     &     theta,pdf_theta)
      call choose_phi_in_interval(phi_min,phi_max,
     &     phi,pdf_phi)
      u0(1)=dcos(theta)*dcos(phi)
      u0(2)=dcos(theta)*dsin(phi)
      u0(3)=dsin(theta)

      return
      end

      

      subroutine choose_theta_in_interval(theta_min,theta_max,
     &     theta,pdf_theta)
      implicit none
      include 'max.inc'
      include 'param.inc'
c
c     Purpose: to choose an angle theta (latitude)
c     uniformally in the [theta_min,theta_max] range
c     
c     Input:
c       + theta_min: lower value of theta in the [-pi/2,pi/2] range
c       + theta_max: higher value of theta in the [-pi/2,pi/2] range
c
c     Output:
c       + theta: value of theta in the [-pi/2,pi/2] range
c       + pdf_theta: associated probability density function
c
c     I/O
      double precision theta_min
      double precision theta_max
      double precision theta
      double precision pdf_theta
c     temp
      double precision R
c     label
      character*(Nchar_mx) label
      label='subroutine choose_theta_in_interval'

      if (theta_min.lt.-pi/2.0D+0) then
         call error(label)
         write(*,*) 'theta_min=',theta_min
         write(*,*) 'is < -pi/2'
         stop
      endif
      if (theta_max.gt.pi/2.0D+0) then
         call error(label)
         write(*,*) 'theta_max=',theta_max
         write(*,*) 'is > pi/2'
         stop
      endif
      if (theta_min.gt.theta_max) then
         call error(label)
         write(*,*) 'theta_min=',theta_min
         write(*,*) 'is > theta_max=',theta_max
         stop
      endif
c
      call random_gen(R)
      theta=dasin(R*(dsin(theta_max)-dsin(theta_min))+dsin(theta_min))
c     
      if ((theta.lt.-pi/2.0D+0).or.(theta.gt.pi/2.0D+0)) then
         call error(label)
         write(*,*) 'theta=',theta
         write(*,*) 'should be in the [-pi/2,pi/2] range'
         stop
      else
         pdf_theta=dcos(theta)/(dsin(theta_max)-dsin(theta_min))
      endif

      return
      end



      subroutine choose_phi_in_interval(phi_min,phi_max,
     &     phi,pdf_phi)
      implicit none
      include 'max.inc'
      include 'param.inc'
c
c     Purpose: to choose an angle phi (longitude)
c     uniformally in the [phi_min,phi_max] range
c     
c     Input:
c       + phi_min: lower value of phi in the [0,2*pi] range
c       + phi_max: higher value of phi in the [0,2*pi] range
c
c     Output:
c       + phi: value of phi in the [0,2*pi] range
c       + pdf_phi: associated probability density function
c
c     I/O
      double precision phi_min
      double precision phi_max
      double precision phi
      double precision pdf_phi
c     temp
      double precision R
c     label
      character*(Nchar_mx) label
      label='subroutine choose_phi_in_interval'

      if (phi_min.gt.phi_max) then
         call error(label)
         write(*,*) 'phi_min=',phi_min
         write(*,*) 'is > phi_max=',phi_max
         stop
      endif
c
      call random_gen(R)
      phi=R*(phi_max-phi_min)+phi_min
c
      pdf_phi=1.0D+0/(phi_max-phi_min)
      
      return
      end



      subroutine choose_scattering_angle(dim,
     &     generate_angle,
     &     pf,g_HG,norm_factor,
     &     alpha,phase_function)
      implicit none
      include 'param.inc'
      include 'max.inc'
      include 'rdgfa.inc'
c
c     Purpose: to return a scattering angle alpha [0,pi]
c
c     Input:
c       + dim: dimension of space
c       + generate_angle: if true, the this routine will generate a new scattering
c         angle ("alpha"). If false, this routine will take "alpha" as an input, and
c         generate the corresponding pdf ("phase_function").
c       + pf: option for choosing the phase function that must be used
c         - pf=0: use the Rayleigh phase function
c         - pf=1: use a isotropic phase function
c         - pf=2: use the Henyey-Greenstein phase function
c         - pf=3: use the RDG-FA phase function
c       + g_HG: Henyey-Greenstein phase-function asymetry factor
c       + norm_factor: normalization factor (so that the integral over 4pi sr is equal to 1)
c       + alpha: if "generate_angle" is false, alpha is an input
c     
c     Output:
c       + alpha: scattering angle [0,pi] is "generate_alpha" is true
c       + phase_function: value of the phase function for the selected angle
c
c     I/O
      integer dim
      logical generate_angle
      integer pf
      double precision g_HG
      double precision norm_factor
      double precision alpha
      double precision phase_function
c     temp
      double precision R
      integer i,angle,j,i0,i1
      double precision mu_icp(1:Nicp_mx)
      double precision icp_min,icp_max
      double precision dtheta
      double precision u,S,mu
      double precision sum
      logical interval_found
      integer interval
      double precision theta,pdf_theta,theta1,theta2
      double precision phi,pdf_phi
      double precision udir(1:Ndim_mx)
      double precision vdir(1:Ndim_mx)
c     label
      character*(Nchar_mx) label
      label='subroutine choose_scattering_angle'
      
      if (generate_angle) then
c     when generating the scattering angle is required, then compute "alpha" from "mu"
         call random_gen(R)
c     
         if (pf.eq.0) then      ! use the Rayleigh phase-function
            u=4.0D+0*R-2.0D+0
            S=(u+dsqrt(1.0D+0+u**2))**(1.0D+0/3.0D+0)
            mu=S-1.0D+0/S
c
         else if (pf.eq.1) then ! use the isotropic phase-function
            call choose_theta_volume(theta,pdf_theta)
            call choose_phi_uniform(phi,pdf_phi)
            udir(1)=dcos(theta)*dcos(phi)
            udir(2)=dcos(theta)*dsin(phi)
            udir(3)=dsin(theta)
            vdir(1)=1.0D+0
            vdir(2)=0.0D+0
            vdir(3)=0.0D+0
            call scalar_product(dim,udir,vdir,mu)
c     
         else if (pf.eq.2) then ! use the Henyey-Greenstein phase-function
            if ((g_HG.lt.-1.0D+0).or.(g_HG.gt.1.0D+0)) then
               call error(label)
               write(*,*) 'Bad value for input argument g_HG=',g_HG
               stop
            endif
            mu=2.D+0*R*(1.D+0+g_HG)**2*(g_HG*(R-1.D+0)+1.D+0)/
     &           (1.D+0-g_HG*(1.D+0-2.D+0*R))**2-1.D+0
c     
         else if (pf.eq.3) then ! use the RDG-FA phase-function
            interval_found=.false.
            do i=1,Nint
               if ((pf_cumulative(i-1).le.R).and.
     &              (pf_cumulative(i).ge.R)) then
                  interval_found=.true.
                  interval=i
                  goto 111
               endif
            enddo               ! interval
 111        continue
            if (.not.interval_found) then
               call error(label)
               write(*,*) 'pf_cumulative interval not found'
               write(*,*) 'R=',R
               write(*,*) 'pf_cumulative(0)=',pf_cumulative(0)
               write(*,*) 'pf_cumulative(',Nint,')=',pf_cumulative(Nint)
               stop
            else
               dtheta=pi/dble(Nint)
               theta1=dtheta*(dble(interval-1))
               theta2=dtheta*(dble(interval))
               theta=theta1
     &              +(theta2-theta1)
     &              *(R-pf_cumulative(interval-1))
     &              /(pf_cumulative(interval)-pf_cumulative(interval-1))
               mu=dcos(theta)
            endif
c     
         endif                  ! pf
c
         alpha=dacos(mu)
      else                      ! generate_angle=F
c     otherwise we just want to know the value of "phase_function" for the provided
c     value of "alpha" -> "mu" has to be recomputed for the provided "alpha"
         mu=dcos(alpha)
      endif                     ! generate_angle
c
      call scattering_phase_function(pf,g_HG,norm_factor,
     &     mu,phase_function)

      return
      end



      subroutine choose_tau_limited(tau_max,tau,pdf_tau)
      implicit none
      include 'max.inc'
c
c     Purpose: to generate an optical depth in the [0,tau_max] range according to:
c     pdf(tau)=exp(-tau)/[1-exp(-tau_max)]
c
c     Input:
c       + tau_max: maximum optical depth
c
c     Output:
c       + tau: random optical depth chosen over the [0,tau_max] range according to pdf(tau)
c       + pdf_tau: value of pdf(tau)
c
c     I/O
      double precision tau_max
      double precision tau
      double precision pdf_tau
c     temp
      double precision R
c     label
      character*(Nchar_mx) label
      label='subroutine choose_tau_limited'

c     security
      if (tau_max.lt.0.0D+0) then
         call error(label)
         write(*,*) 'tau_max=',tau_max
         stop
      endif

c     special case tau_max=0
      if (tau_max.eq.0.0D+0) then
         tau=0.0D+0
         pdf_tau=1.0D+0
c     general case tau_max>0
      else
         call random_gen(R)
         tau=-dlog(1.0D+0-R*(1.0D+0-dexp(-tau_max)))
         pdf_tau=dexp(-tau)/(1.0D+0-dexp(-tau_max))
      endif

      return
      end



      subroutine choose_tau_unlimited(tau,pdf_tau)
      implicit none
      include 'max.inc'
c
c     Purpose: to generate an optical depth in the [0,+infty) range according to:
c     pdf(tau)=exp(-tau)
c
c     Output:
c       + tau: random optical depth chosen over the [0,+infty) range according to pdf(tau)
c       + pdf_tau: value of pdf(tau)
c
c     I/O
      double precision tau
      double precision pdf_tau
c     temp
      double precision r
c     label
      character*(Nchar_mx) label
      label='subroutine choose_tau_unlimited'

      call random_gen(r)
      tau=-dlog(r)
      pdf_tau=dexp(-tau)

      return
      end



      subroutine choose_length_unlimited(k,length,pdf_l)
      implicit none
      include 'max.inc'
c
c     Purpose: to generate a length in the [0,+infty) range according to:
c     pdf(length)=k*exp(-k*length)
c
c     Input:
c       + k: optical coefficient (absorption, scattering or total extinction) (m^-1)
c
c     Output:
c       + length: random length chosen over the [0,+infty) range according to pdf(length)
c       + pdf_l: value of pdf(length)
c
c     I/O
      double precision k
      double precision length
      double precision pdf_l
c     temp
      double precision R
c     label
      character*(Nchar_mx) label
      label='subroutine choose_length_unlimited'

      if (k.lt.0.0D+0) then
         call error(label)
         write(*,*) 'Bad input k=',k
         stop
      else if (k.eq.0.0D+0) then
         length=1.0D+300
         pdf_l=0.0D+0
      else
         call random_gen(r)
         length=-1.0D+0/k*dlog(r)
         pdf_l=k*dexp(-k*length)
      endif

      return
      end
