c     Copyright (C)
c     2021 Mésotar
c     
c     This file must be used under the terms of the CeCILL license.
c     This source file is licensed as described in the file COPYING, which
c     you should have received as part of this distribution.  The terms
c     are also available at
c     http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.txt
c
      subroutine sw_direct(dim,box_volume,laser_volume,
     &     emission_surface,integration_surface,
     &     d,phi0,nc_tr_version,pf,g_HG,norm_factor,Nevent)
      implicit none
      include 'max.inc'
      include 'com_properties.inc'
c     
c     Purpose: to perform the computation of the flux incoming on the
c     integration surface using a direct MC algorithm
c     
c     Input:
c       + dim: dimension of space
c       + box_volume: volume limits of the surrounding box
c       + laser_volume: volume limits of the laser sheet
c       + emission_surface: surface limits of the emission surface
c       + integration_surface: surface limits of the integration surface
c       + d: distance between laser sheet and integration surface [m]
c       + phi0: Laser surface density power [W/m²]
c       + nc_tr_version: version of null-collision for evaluating the transmissivity
c       + pf: choice of phase function [0: Rayleigh, 1:isotropic, 2: Henyey-Greenstein, 3: RDG-FA]
c       + g_HG: value of the asymetry parameter when using a Henyey-Greenstein phase function
c       + norm_factor: normalization factor (so that the integral over 4pi sr is equal to 1)
c       + Nevent: number of statistical events
c     
c     Output:
c       + on-screen: value of the flux [W]
c     
c     I/O
      integer dim
      double precision box_volume(1:Ndim_mx,1:2)
      double precision laser_volume(1:Ndim_mx,1:2)
      double precision emission_surface(1:Ndim_mx,1:2)
      double precision integration_surface(1:Ndim_mx,1:2)
      double precision d
      double precision phi0
      integer nc_tr_version
      integer pf
      double precision g_HG
      double precision norm_factor
      integer Nevent
c     temp
      integer i,j,event,ivar
      logical debug,check1
      double precision x0(1:Ndim_mx)
      double precision x1(1:Ndim_mx)
      double precision u0(1:Ndim_mx)
      double precision u1(1:Ndim_mx)
      double precision n0(1:Ndim_mx)
      double precision ks
      double precision S0,pdf_u0,Sdetector
      logical intersection_found
      double precision x_int(1:Ndim_mx)
      double precision d_x0xint
      double precision n_int(1:Ndim_mx)
      integer errcode
      integer Nvar
      double precision sum(1:Nvar_mx)
      double precision sum2(1:Nvar_mx)
      double precision contrib(1:Nvar_mx),weight,w
      double precision mean(1:Nvar_mx)
      double precision variance(1:Nvar_mx)
      double precision std_dev(1:Nvar_mx)
      double precision r
      logical on_surface
      integer surface_index1,surface_index2
      double precision prod,tau,pdf_tau
      logical keep_going,x1_is_in_box
      double precision length
      logical generate_angle
      double precision alpha,phase_function
c     progress display
      integer len,ndone,ntot
      double precision fdone,fdone0
      integer ifdone,pifdone
      character*(Nchar_mx) str,source_str
      character*(Nchar_mx) fmt,fmt0
      integer err
c     label
      character*(Nchar_mx) label
      label='subroutine sw_direct'

      Nvar=1
c     ----------------- print status
      len=6
      call num2str(len,str,err)
      if (err.eq.0) then
         fmt0='(a,i'//trim(str)//',a)'
         fmt='(i'//trim(str)//',a)'
      endif
c     ----------------- print status
c     check
      if (Nvar.gt.Nvar_mx) then
         call error(label)
         write(*,*) 'Nvar=',Nvar
         write(*,*) '> Nvar_mx=',Nvar_mx
         stop
      endif
c     Initialization
      do ivar=1,Nvar
         sum(ivar)=0.0D+0
         sum2(ivar)=0.0D+0
      enddo ! ivar
      
c     ----------------- print status
      write(*,*) 'Performing direct MC computation...'
      write(*,trim(fmt0),advance='no')
     &     'Done:   ',floor(fdone0),' %'
      ndone=0
      ntot=Nevent
      pifdone=0
c     ----------------- print status

c     emission surface
      S0=(emission_surface(1,2)-emission_surface(1,1))*
     &     (emission_surface(3,2)-emission_surface(3,1)) ! [m²]
c     integration surface
      Sdetector=(integration_surface(1,2)-integration_surface(1,1))*
     &     (integration_surface(2,2)-integration_surface(2,1))
c     normal to the emission surface
      n0(1)=0.0D+0
      n0(2)=1.0D+0
      n0(3)=0.0D+0

      do event=1,Nevent
c     weight initialization
         do ivar=1,Nvar
            contrib(ivar)=0.0D+0
         enddo                  ! ivar
         weight=phi0            ! [W/m²]
c     sampling the emission surface
         do i=1,dim
            call random_gen(r)
            x0(i)=emission_surface(i,1)
     &           +(emission_surface(i,2)-emission_surface(i,1))*r
         enddo                  ! i
c     weight=weight/pdf(S) with pdf(S)=1/S0
         weight=weight*S0       ! [W]
c     check
         call is_on_surface(dim,x0,laser_volume,
     &        on_surface,surface_index1,surface_index2)
         if (.not.on_surface) then
            call error(label)
            write(*,*) 'x0=',x0
            write(*,*) 'is not on the surface of laser_volume'
            stop
         endif
c     emission direction: u0
         call duplicate_vector(dim,n0,u0)
c     
         keep_going=.true.
         do while (keep_going)
c     Sampling scattering distance
            call choose_tau_unlimited(tau,pdf_tau)
            length=tau/ks_hat
            call direct_flight(dim,x0,u0,length,x1)
c     check x1 in or out surrounding box
            call is_in_box(dim,x1,box_volume,x1_is_in_box)
            if (x1_is_in_box) then
c     update the weight
               call null_collision_Tr(dim,box_volume,
     &              nc_tr_version,x0,x1,'abs',w)
               weight=weight*w
               call scattering_coefficient(dim,box_volume,x1,ks)
               call random_gen(r)
               if (r.le.ks/ks_hat) then ! scattering event
c     Scattering @ x1
                  generate_angle=.true. ! generate scattering angle
                  call choose_scattering_angle(dim,
     &                 generate_angle,
     &                 pf,g_HG,norm_factor,
     &                 alpha,phase_function)
                  call direction_after_scattering(dim,u0,alpha,u1)
               else             ! null-collision
                  call duplicate_vector(dim,u0,u1)
               endif
               call duplicate_vector(dim,x1,x0)
               call duplicate_vector(dim,u1,u0)
               if (weight.le.0.0D+0) then
                  keep_going=.false.
               endif
            else                ! x1_is_in_box=F
c     Identify intersection with surrounding box
               debug=.false.
               check1=.true.
               call intersection_with_box_from_inside(debug,dim,check1,
     &              box_volume,x0,u0,
     &              intersection_found,x_int,d_x0xint,n_int,errcode)
               if (.not.intersection_found) then
                  call error(label)
                  write(*,*) 'Intersection not found with '
     &                 //'box_volume from inside using:'
                  write(*,*) 'x0=',x0
                  write(*,*) 'u0=',u0
                  stop
               endif            ! intersection_found=F
c     transmissivity, and update of weight
               call null_collision_Tr(dim,box_volume,nc_tr_version,
     &              x0,x_int,'abs',w)
               weight=weight*w
c     Check if x_int is on integration surface
               if ((x_int(1).ge.integration_surface(1,1)).and.
     &              (x_int(1).le.integration_surface(1,2)).and.
     &              (x_int(2).ge.integration_surface(2,1)).and.
     &              (x_int(2).le.integration_surface(2,2)).and.
     &              (dabs(x_int(3)-integration_surface(3,1))
     &              .lt.1.0D-6)) then
c     If the intersection position is located on the integration surface,
c     the contribution of the path is equal to the current weight, and
c     the realization ends
                  contrib(1)=weight
                  keep_going=.false.
               else
c     otherwise, the contribution of the path is null, and the realization ends
                  contrib(1)=0.0D+0
                  keep_going=.false.
               endif
            endif               ! x1_is_in_box
         enddo                  ! while (keep_going)
c     update of sum and sum2
         do ivar=1,Nvar
            sum(ivar)=sum(ivar)+contrib(ivar)
            sum2(ivar)=sum2(ivar)+contrib(ivar)**2.0D+0
         enddo                  ! ivar
c     ----------------- print status
         ndone=ndone+1
         fdone=dble(ndone)/dble(ntot)*1.0D+2
         ifdone=floor(fdone)
         if (ifdone.gt.pifdone) then
            do j=1,len+2
               write(*,"(a)",advance='no') "\b"
            enddo               ! j
            write(*,trim(fmt),advance='no')
     &           floor(fdone),' %'
            pifdone=ifdone
         endif
c     ----------------- print status
      enddo                     ! event

      do ivar=1,Nvar
         call statistics(Nevent,sum(ivar),sum2(ivar),
     &        mean(ivar),variance(ivar),std_dev(ivar))
      enddo                     ! ivar
      write(*,*)
      write(*,*) 'Direct MC result:'
      write(*,*) 'Absorbed Flux=',mean(1),' +/-',std_dev(1),' W'
      write(*,*) 'Absorbed Flux density=',mean(1)/Sdetector,
     &     ' +/-',std_dev(1)/Sdetector,' W/m²'
      write(*,*) 'Pa/Pe=',mean(1)/(phi0*S0),' +/-',std_dev(1)/(phi0*S0)
      
      return
      end
